////////////////////////////////////////////////////////////////////////
// Class:       FEBOverflowTrigger
// Module Type: filter
// File:        FEBOverflowTrigger_module.cc
//
// Generated at Mon Apr 14 17:00:29 2014 by Luke Vinton using artmod
// from cetpkgsupport v1_05_03.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"

//Copied from HighETrigger
#include "art/Persistency/Common/Assns.h"
#include "DDTUtilities/AssociationUtil.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"

#include <memory>
//copied from HighETrigger
#include <string>
#include <algorithm>

namespace novaddt {
  class FEBOverflowTrigger;
}

class novaddt::FEBOverflowTrigger : public art::EDFilter {
public:
  explicit FEBOverflowTrigger(fhicl::ParameterSet const & p);
  virtual ~FEBOverflowTrigger();

  bool filter(art::Event & e) override;


private:
  std::string hit_label_, hit_instance_;
  unsigned prescale_, trigger_counts_;

};


novaddt::FEBOverflowTrigger::FEBOverflowTrigger(fhicl::ParameterSet const & p)
  :
  hit_label_(p.get<std::string>("hit_label")),
  hit_instance_(p.get<std::string>("hit_instance")),
  prescale_(p.get<unsigned>("prescale")),
  trigger_counts_(0)
{
  produces<std::vector<TriggerDecision>>();
}

novaddt::FEBOverflowTrigger::~FEBOverflowTrigger()
{
  std::cout << "\n\tNumber of FEB overflow Triggers = " << trigger_counts_
	    << "\n" << std::endl;
}

bool novaddt::FEBOverflowTrigger::filter(art::Event & e)
{
  bool result = false;
  std::unique_ptr<std::vector<TriggerDecision>> 
    trigger_decisions(new std::vector<TriggerDecision>);

  art::Handle<novaddt::HitList > hits;
  e.getByLabel(hit_label_, hit_instance_, hits);
  
  bool foundOverflowHit = false;
  novaddt::TDC overflowTime = 0;
  
  for (auto const& hit : *hits){
    if(hit.FEBOverflowError()){
      foundOverflowHit=true;
      overflowTime = hit.TDC();
    }
  }
  
  //TDC dt = 64000; //64MHz * 1 ms
  //TDC s = 0.95;
  //TDC t0 = overflowTime - 0.95 * dt; // start time is 950 us back from overflow

  double dt = 64000;
  double sdt = 0.95 * dt;
  TDC tdt = sdt;
  TDC t0 = overflowTime - tdt; // start time is 950 us back from overflow
  
  if(foundOverflowHit){
    ++trigger_counts_;
    if (trigger_counts_ % prescale_ == prescale_ - 1) 
      {
	trigger_decisions->emplace_back
	  (t0, dt, daqdataformats::trigID::TRIG_ID_DATA_FEBOVERFLOW, prescale_);
	result = true;
      }
  }
    
  e.put(std::move(trigger_decisions));
  return result;
  
}

DEFINE_ART_MODULE(novaddt::FEBOverflowTrigger)
