#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/HoughPoint.h"

#include <TVector2.h>

#include <cmath>
#include <cstdio>

namespace novaddt{

  //------------------------------------------------------------------------------
  // Default constructor
  HoughPoint::HoughPoint():
    _p1(nullptr),
    _coord1(nullptr),
    _p2(nullptr),
    _coord2(nullptr),
    _cosv(0.0f),
    _doca(0.0f),
    _vpro(0.0f)
  {
  }

  //------------------------------------------------------------------------------
  HoughPoint::HoughPoint(const DAQHit& p1, const TVector2& c1, const DAQHit& p2, const TVector2& c2):
    _p1(&p1),
    _coord1(&c1),
    _p2(&p2),
    _coord2(&c2),
    _cosv(0.0f),
    _doca(0.0f),
    _vpro(0.0f)
  {
    // Compute the hough parameters
    calc();
  }

  void HoughPoint::reset(const DAQHit& p1, const TVector2& c1, const DAQHit& p2, const TVector2& c2)
  {
    _p1     = &p1;
    _coord1 = &c1;
    _p2     = &p2;
    _coord2 = &c2;
    calc();
  }

  void HoughPoint::reset(){
    _p1 = nullptr;
    _coord1 = nullptr;
    _p2 = nullptr;
    _coord2 = nullptr;
    _cosv = 0.0f;
    _doca = 0.0f;
    _vpro = 0.0f;
  }

  void HoughPoint::print() const {
    // Print the transform's record
    printf("Cosv: %2.4f\tDoca: %2.4f\t\nHits:\n",
	   _cosv,
	   _doca
	   );
    _p1->print();
    _p2->print();
  }

  void HoughPoint::calc(){

    // As efficient as possible: no unnecessary copies.
    /*
      float x1(3.359*(_p1->Plane().val-480));
      float x2(3.359*(_p2->Plane().val-480));
      float y1(_p1->Cell().val-191.5);
      float y2(_p2->Cell().val-191.5);
    */
    uint64_t t1(_p1->TDC().val);
    uint64_t t2(_p2->TDC().val);
    /*
    //If same chits are provided: assert false
    if (coord1.X()==x2 && y1==y2)
    {
    std::cerr<<"Same cells have been provided!"<<std::endl;
    assert(false);
    }
    */

    _dist = (*_coord1-*_coord2).Mod();

    _cosv = (_coord2->Y()-_coord1->Y())/_dist;
    _doca = (_coord2->X()*_coord1->Y() - _coord1->X()*_coord2->Y())/_dist;
  
    if (t1>t2) _vpro = (t1-t2)/_dist;
    else _vpro = (t2-t1)/_dist;

    if (_coord1->X()>_coord2->X()) {
      _cosv = -_cosv;
      _doca = -_doca;
      _vpro = -_vpro;
    }

  }
} // namespace
