////////////////////////////////////////////////////////////////////////////
// \version $Id: Track3D.cxx,v 1.0 2013-06-25 tamsett
//
// \brief Definition of Track3D object
//
// \author tamsett@fnal.gov
//
////////////////////////////////////////////////////////////////////////////
#ifndef NOVADDT_TRACK3D_H
#define NOVADDT_TRACK3D_H

#include <DDTBaseDataProducts/BaseProducts.h>
#include "TVector3.h"
//---------------------------------------------------------------------
namespace novaddt {
  class Track3D{
  public:
    Track3D();
  private:
    bool     fIs3D;        // true if track is in both xz & yz views
    int      fView;        // 0: yz, 1: xz, 2: xyz
    TVector3 fStart;       // track start
    TVector3 fEnd;         // track end
#ifndef __GCCXML__
  public:
    explicit Track3D(
        bool     const& is3D,
        int      const& view,
        TVector3 const& start,
        TVector3 const& end
    );
    bool     const& Is3D() const;
    int      const& View() const;
    TVector3 const& Start() const;
    TVector3 const& End() const;
#endif
  };
#ifndef __GCCXML__
  inline bool     const& Track3D::Is3D()   const { return fIs3D;  }
  inline int      const& Track3D::View()   const { return fView;  }
  inline TVector3 const& Track3D::Start()  const { return fStart; }
  inline TVector3 const& Track3D::End()    const { return fEnd;   }  
#endif 
}
//---------------------------------------------------------------------
#endif //NOVADDT_TRACK3D_H
