////////////////////////////////////////////////////////////////////////
// Class:       GHLTest
// Module Type: analyzer
// File:        GHLTest_module.cc
// /brief       Make fake DAQHits to test usage of GroupedHitLists
//
// Generated at Tue Sep 18 13:03:22 2012 by Gavin Davies using artmod
// from art v1_01_01.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/GroupedHitList.h"
#include "DDTBaseDataProducts/HitList.h"


namespace novaddt {
  class GHLTest;
}


class novaddt::GHLTest : public art::EDAnalyzer {
public:
  explicit GHLTest(fhicl::ParameterSet const & p);
  virtual ~GHLTest();

  virtual void analyze(art::Event const & e);

  virtual void beginJob();
  virtual void endJob();

private:

  void print(DAQHit const& h) const;
};



novaddt::GHLTest::GHLTest(fhicl::ParameterSet const & p) :
  EDAnalyzer(p)
{
}



novaddt::GHLTest::~GHLTest()
{
  // Clean up dynamic memory and other resources here.
}



void novaddt::GHLTest::print(DAQHit const& h) const
{
  std::cout << "TDC: " << h.TDC().val 
	    <<", ADC: " << h.ADC().val
	    <<", Plane: " << h.Plane().val
	    <<std::endl;
}



void novaddt::GHLTest::analyze(art::Event const & /*e*/)
{
  // Let's create a fake list of DAQ hits:  
  HitList dhits;
  dhits.
    push_back(DAQHit(novaddt::TDC(42),novaddt::ADC(600),novaddt::Plane(27)));  
  dhits.
    push_back(DAQHit(novaddt::TDC(50),novaddt::ADC(300),novaddt::Plane(20)));  
  dhits.
    push_back(DAQHit(novaddt::TDC(90),novaddt::ADC(500),novaddt::Plane(22)));  
  dhits.
    push_back(DAQHit(novaddt::TDC(70),novaddt::ADC(400),novaddt::Plane(21)));  
  dhits.
    push_back(DAQHit(novaddt::TDC(63),novaddt::ADC(350),novaddt::Plane(12)));  


  // Here is how we generate a grouped hit list:
  // 1. sort the hits (done by HitSorters)
  std::sort(dhits.begin(), dhits.end(), CompareDAQHit<TDC>());

  // 2. create boundary list (done by HitSlicers)
  std::vector<std::size_t> boundaries = {0, 2, dhits.size()};
  GroupedHitList<TDC> glist(dhits.begin(), dhits.end(), boundaries);

  
  // Now, let's take a look at our grouped hit list:
  std::cout << "\nAll Hits." << std::endl;
  for(auto const & hit : dhits)
    print(hit);

  std::cout << "\nSorted Groups." << std::endl;
  for (unsigned group = 0; group != glist.n_groups(); ++group)
  {
    std::cout << "\nGroup " << group 
	      << " (size = " << glist.group_size(group) << "):" 
	      << std::endl;

    // extract one group and loop over it:
    HitList hit_group = glist.get(group);
    for (auto const & h : hit_group)
      print(h);
  }

  std::cout << "\nWhole Sorted List." << std::endl;
  HitList new_list = glist.get();
  for (auto const & n : new_list)
    print(n);

  std::cout << "First Element: " << glist.first_value()
	    << "\n Last Element: " << glist.last_value()
	    << std::endl;
}



void novaddt::GHLTest::beginJob()
{
  // Implementation of optional member function here.
}



void novaddt::GHLTest::endJob()
{
  // Implementation of optional member function here.
}



DEFINE_ART_MODULE(novaddt::GHLTest)
