////////////////////////////////////////////////////////////////////////
// Class:       SlicePrescale
// Module Type: filter
// File:        SlicePrescale_module.cc
//
// Generated at Thu Feb 13 05:14:01 2014 by Matthew Tamsett using artmod
// from cetpkgsupport v1_04_02.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"

#include <iostream>
#include <memory>

#include <stdlib.h> // for rand
#include <time.h>   // for rand seed
#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"

//---------------------------------------------------------------------
namespace novaddt {
    class SlicePrescale;
}
//---------------------------------------------------------------------
class novaddt::SlicePrescale : public art::EDFilter {
    public:
        explicit SlicePrescale(fhicl::ParameterSet const & p);
        virtual ~SlicePrescale();
        bool filter(art::Event & e) override;
        void endJob() override;
    private:
        // Declare member data here.
        // Configuration options
        unsigned int _prescale;           ///< prescale factor (1 out of every this many passes are issued as real triggers)
        std::string  _slice_label;        ///< label of module making slices
        std::string  _slice_instance;     ///< instance of module making slices
        std::string  _mode;               ///< prescale mode, can be either fixed or random
        unsigned int _seed;               ///< random number generator seed
        // counters
        unsigned int _nEvents           = 0;
        unsigned int _nSlices           = 0;
        unsigned int _nAccepted         = 0;
        
};
//---------------------------------------------------------------------
novaddt::SlicePrescale::SlicePrescale(fhicl::ParameterSet const & p)
:   _prescale          (p.get<unsigned int>("prescale"              ))
,   _slice_label       (p.get<std::string >("slice_label"           ))
,   _slice_instance    (p.get<std::string >("slice_instance"        ))
,   _mode              (p.get<std::string >("mode"                  ))
,   _seed              (p.get<unsigned int>("seed",       time(NULL)))
{
    std::cout << "--- novaddt::SlicePrescale instantiate" << std::endl;
    std::cout << "\t prescale:                      " << _prescale      << std::endl;
    std::cout << "\t slice label:                   " << _slice_label   << std::endl;
    std::cout << "\t slice instance:                " << _slice_instance<< std::endl;
    std::cout << "\t mode:                          " << _mode          << std::endl;
    std::cout << "\t seed:                          " << _seed          << std::endl;
    
    // assert that the mode requested is configured
    assert( (_mode == "fixed") || (_mode == "random") );
    
    // seed the random number generator, use a random default if not specified    
    // in the configuration file.  
    srand(_seed);
    
    // Call appropriate Produces<>() functions here.    
    produces<std::vector<HitList>>("Prescaled"+_slice_instance);
}
//---------------------------------------------------------------------
novaddt::SlicePrescale::~SlicePrescale()
{
    // Clean up dynamic memory and other resources here.
}
//---------------------------------------------------------------------
bool novaddt::SlicePrescale::filter(art::Event & event)
{
    LOG_DEBUG("SlicePrescale")<< "--- novaddt::SlicePrescale filter. Event: " 
              << event.id().event() 
              << std::endl;
    _nEvents++;
    
    // make output product
    std::unique_ptr< std::vector<HitList> >product(new std::vector<HitList>);
    
    // get input product
    art::Handle< std::vector<novaddt::HitList> > slices;
    event.getByLabel(_slice_label, _slice_instance, slices);
    LOG_DEBUG("SlicePrescale") << "\t - got " << slices->size() << " slices" << std::endl;
    
    // loop over slices
    for(size_t i = 0; i < slices->size(); ++i){
        _nSlices++;
        // check prescale
        if ( ((_mode=="fixed")  && ( _nSlices%_prescale == _prescale-1)) ||
             ((_mode=="random") && (rand()%_prescale    == _prescale-1)) ){
                _nAccepted++;
                LOG_DEBUG("SlicePrescale")<< "\t - not prescaled" << std::endl;
                // add this hit list into our output product
                HitList slice = slices->at(i);
                product->push_back(slice);
        } // end of check on prescale
    } // end of loop over slices
    
    // put the hits into the event for later use
    event.put(std::move(product),"Prescaled"+_slice_instance);
  
    // always return true
    return true;
}
//---------------------------------------------------------------------
void novaddt::SlicePrescale::endJob()
{
    std::cout << "--- novaddt::SlicePrescale endJob" << std::endl;
    std::cout << "\tNumber of events seen:         " << _nEvents             << std::endl;
    std::cout << "\tNumber of slices seen:         " << _nSlices             << std::endl;
    std::cout << "\tNumber of slices accepted:     " << _nAccepted           << std::endl;
}
//---------------------------------------------------------------------
DEFINE_ART_MODULE(novaddt::SlicePrescale)
