////////////////////////////////////////////////////////////////////////
// Class:       SingletonRejectionAna
// Module Type: analyzer
// File:        SingletonRejectionAna_module.cc
//
// Generated at Tues June 26 2013 by Will Henderson	
// Based on NuETriggerAna and MonopoleAna
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Principal/Event.h"

// NOvADDT includes
#include "DDTBaseDataProducts/Cluster.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/BaseProducts.h"

// ROOT includes
#include "TH1F.h"
#include "TH2F.h"

namespace novaddt {
  class SingletonRejectionAna;
}
///////////////////////////////////////////////////////////////////////
class novaddt::SingletonRejectionAna : public art::EDAnalyzer {
  
public:
  explicit SingletonRejectionAna(fhicl::ParameterSet const & p);
  virtual ~SingletonRejectionAna();
  void analyze(art::Event const & e) override;
  void beginJob() override;
  void endJob() override;
  
private:
  // Declare member data
  std::string _ModuleLabel; 
  std::string _InstanceLabel; 
  TH1F* rawADC;
  TH2F* HitsYZ; // hit locations in YZ view	
  TH2F* HitsXZ; // hit locations in XZ view
};
///////////////////////////////////////////////////////////////////////
novaddt::SingletonRejectionAna::SingletonRejectionAna(fhicl::ParameterSet const & p)
  : EDAnalyzer(p),
    _ModuleLabel(p.get<std::string>("ModuleLabel")),
    _InstanceLabel(p.get<std::string>("InstanceLabel"))
{
  //Initialize member data
}
///////////////////////////////////////////////////////////////////////
novaddt::SingletonRejectionAna::~SingletonRejectionAna()
{
  // Clean up dynamic memory and other resources
}
///////////////////////////////////////////////////////////////////////
void novaddt::SingletonRejectionAna::analyze(art::Event const & e)
{
  // Implementation of member functions
  mf::LogInfo("SingletonRejectionAna") << "=== novaddt::SingletonRejectionAna Event: " << e.id().event() << " ===" << std::endl;
  
  // Input grouped hitlists
  art::Handle< std::vector<novaddt::HitList> > hit_lists;
  e.getByLabel(_ModuleLabel, _InstanceLabel, hit_lists);
  
  /*
  // Check for empty GroupedHitList, HitList, or Track
  if(hit_lists->empty()){
  mf::LogWarning ("No HitLists")<<"No HitLists in the input file"; 
  return;
  }
  */  
  
  for(unsigned int i=0; i < hit_lists->size(); ++i){ // Loop of hitlists
    
    HitList hits = hit_lists->at(i);
    
    // Loop over hits
    for(unsigned int j=0; j < hits.size();++j){
      rawADC -> Fill(hits[j].ADC().val);
      
      if (!(hits[j].View().val==daqchannelmap::Y_VIEW)){
	//XZ projection
	HitsXZ -> Fill(hits[j].Plane().val, hits[j].Cell().val);
      } else{
	//YZ projection
	HitsYZ -> Fill(hits[j].Plane().val, hits[j].Cell().val);
      }      
      
    } //END j
  } //END i
  
  /*  
  // hitProducer (unfiltered, unsliced hitlist example using pointers)
  art::Handle<novaddt::HitList> hits;
  evt.getByLabel(hits_label_, hits);
  std::unique_ptr<HitList> product(new HitList(hits->begin(),hits->end()));
  
  auto begin = product->begin();
  auto end = product->end();
  for (auto i = begin; i!=end; ++i)
  {   
  rawADC -> Fill(i->ADC().val);
  */
  
  return;  
}
/////////////////////////////////////////////////////////////////////// 
void novaddt::SingletonRejectionAna::beginJob()
{
  art::ServiceHandle<art::TFileService> f;
  
  rawADC = f->make<TH1F>("rawADC","Raw ADC",4096,-0.5,4095.5);
  HitsYZ = f->make<TH2F>("hitsYZ", ";z (cm); y (cm)", 1000, 0., 1000., 200, 0., 200.); 
  HitsXZ = f->make<TH2F>("hitsXZ", ";z (cm); x (cm)", 1000, 0., 1000., 200, 0., 200.);   
}
/////////////////////////////////////////////////////////////////////// 
void novaddt::SingletonRejectionAna::endJob()
{
  // Implementation of optional member function
}
/////////////////////////////////////////////////////////////////////// 
DEFINE_ART_MODULE(novaddt::SingletonRejectionAna)
