//////////////////////////////////////////////////////////////////////
// Class:       UpMuTestTrigger
// Module Type: filter
// File:        UpMuTestTrigger_module.cc
//
// Generated at Tues July 17 2013 by Will Henderson
// Based on NuMuTestTrigger (Matthew Tamsett)
//////////////////////////////////////////////////////////////////////

// NOvAART includes
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/FindOneP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Persistency/Common/Assns.h"

// NOvADDT includes
#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/Track.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"

// NOvADAQ includes
#include "DAQChannelMap/DAQChannelMap.h"
#include "DDTUtilities/AssociationUtil.h"

// ROOT includes
#include "TVector2.h"
#include "TVector3.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TFile.h"

//////////////////////////////////////////////////////////////////////
namespace novaddt {
    class UpMuTestTrigger;
}
//////////////////////////////////////////////////////////////////////
class novaddt::UpMuTestTrigger : public art::EDFilter {
public:
    explicit UpMuTestTrigger(fhicl::ParameterSet const & p);
    virtual ~UpMuTestTrigger();
    bool filter(art::Event & e) override;
    void endJob() override;
private:
    unsigned int _prescale;           ///< prescale factor (1 out of every this many passes are issued as real triggers)
    std::string  _TrackModuleLabel;   ///< label of module making the Tracks
    std::string  _TrackInstanceLabel; ///< instance label making the Tracks
    std::string  _SliceModuleLabel;   ///< label of module making the HitList
    std::string  _SliceInstanceLabel; ///< instance label making the HitList
    
    // Minimum track hit cuts
    unsigned int fMinPlanes;          // minimum planes crossed (Z)
    unsigned int fMinCellsX;          // minimum number of cells in X view
    unsigned int fMinCellsY;          // minimum number of cells in Y view
    
    // Containment cuts
    float fMinX;                      // lower x fiducial cut
    float fMaxX;                      // upper x fiducial cut
    float fMinY;                      // lower y fiducial cut
    float fMaxY;                      // upper y fiducial cut
    unsigned int fMinZ;               // lower z fiducial cut
    unsigned int fMaxZ;               // upper z fiducial cut
    
    // Angle cut
    double fAngleCut;                 // accept 30 deg vertical cone (default-- cos angle is angle from horizontal)
    
    // Direction cut
    double fDirectionCut;       // accept upward slope
    
    // Counters
    unsigned int _nEvents               =0;
    unsigned int _nTracks               =0;
    unsigned int _nPass_dhorizontal     =0;
    unsigned int _nPass_dvertical       =0;
    unsigned int _nPass_through         =0;
    unsigned int _nPass_stopped         =0;
    unsigned int _nPass_fully           =0;
    unsigned int _nPass_angle           =0;
    unsigned int _nPass_direction       =0;
    unsigned int _trigger_counts        = 0;
    unsigned int _prescaled_trigger_counts = 0;
    
    // Switches
    bool         fcheckMinHits;       // boolean switch for cut on minimum track hits
    bool         fcheckContainment;   // boolean switch for cut on containment
    bool         fcheckAngle;         // boolean switch for cut on angle
    bool         fcheckDirection;     // boolean switch for cut on direction
    
    // Histograms
    TH2F *HitsXZ = new TH2F("hitsXZ", ";z (plane); x (vertical cell)", 900, 0., 900., 400, 0., 400.); // hit locations in XZ view
    TH2F *HitsYZ = new TH2F("hitsYZ", ";z (plane); y (horizontal cell)", 900, 0., 900., 400, 0., 400.); // hit locations in YZ view
    
    TH1F *fslope = new TH1F("slope",";Slope", 220, -55, 55);                                  // slope of the fit for trigger
    TH1F *fexpected = new TH1F("expected",";expTime (ns)", 80, -40, 40);                      // expected hit time difference
    TH1F *fmeasured = new TH1F("measured", ";meaTime (ns)", 2000, -1000, 1000);                 // measured hit time difference
    
    TH2F *fdirslp = new TH2F("dirslp", ";angle (degrees); slope", 220, -55, 55, 100, 0, 180); // graph of slope vs. angle cos(Y) in the x-z plane
    TH2F *fexpmea = new TH2F("expmea", ";expected time (ns); measured time (ns)", 200, 0, 120, 400, -500, 500); // graph of expected time vs. measured time
    
    /*
     TH1D *ftrue;           // true hit time
     TH1D *fsmear;          // difference between smeared time and true time
     TH2D *fupmuper;        // percent of upmus recognized vs. cut hits in the x-z plane
     */
};
//////////////////////////////////////////////////////////////////////
novaddt::UpMuTestTrigger::UpMuTestTrigger(fhicl::ParameterSet const & p)
: _prescale             (p.get<unsigned int>("prescale"              ))
, _TrackModuleLabel     (p.get<std::string >("TrackModuleLabel"      ))
, _TrackInstanceLabel   (p.get<std::string >("TrackInstanceLabel", ""))
, _SliceModuleLabel     (p.get<std::string >("SliceModuleLabel"      ))
, _SliceInstanceLabel   (p.get<std::string >("SliceInstanceLabel", ""))
, fMinPlanes            (p.get<unsigned int>("MinPlanes"             ))
, fMinCellsX            (p.get<unsigned int>("MinCellsX"             ))
, fMinCellsY            (p.get<unsigned int>("MinCellsY"             ))
, fMinX                 (p.get<float>       ("Xmin"                  ))
, fMaxX                 (p.get<float>       ("Xmax"                  ))
, fMinY                 (p.get<float>       ("Ymin"                  ))
, fMaxY                 (p.get<float>       ("Ymax"                  ))
, fMinZ                 (p.get<unsigned int>("Zmin"                  ))
, fMaxZ                 (p.get<unsigned int>("Zmax"                  ))
, fAngleCut             (p.get<double>      ("AngleCut"              ))
, fDirectionCut         (p.get<double>      ("DirectionCut"          ))
, fcheckMinHits         (p.get<bool >       ("CheckMinHits"          ))
, fcheckContainment     (p.get<bool >       ("CheckContainment"      ))
, fcheckAngle           (p.get<bool >       ("CheckAngle"            ))
, fcheckDirection       (p.get<bool >       ("CheckDirection"        ))
{
    std::cout << "=== novaddt::UpMuTestTrigger begin ==="                   << std::endl;
    std::cout << "\t prescale:                  " << _prescale              << std::endl;
    std::cout << "\t TrackModuleLabel:          " << _TrackModuleLabel      << std::endl;
    std::cout << "\t TrackInstanceLabel:        " << _TrackInstanceLabel    << std::endl;
    std::cout << "\t SliceModuleLabel:          " << _SliceModuleLabel      << std::endl;
    std::cout << "\t SliceInstanceLabel:        " << _SliceInstanceLabel    << std::endl;
    std::cout << "\t MinPlanes:                 " << fMinPlanes             << std::endl;
    std::cout << "\t MinCellsX:                 " << fMinCellsX             << std::endl;
    std::cout << "\t MinCellsY:                 " << fMinCellsY             << std::endl;
    std::cout << "\t X fiducial cuts:           " << fMinX << " to " << fMaxX << std::endl;
    std::cout << "\t Y fiducial cuts:           " << fMinY << " to " << fMaxY << std::endl;
    std::cout << "\t Z fiducial cuts:           " << fMinZ << " to " << fMaxZ << std::endl;
    std::cout << "\t Angle cut:                 " << fAngleCut              << std::endl;
    std::cout << "\t Direction cut:             " << fDirectionCut          << std::endl << std::endl;
    std::cout << "\t CheckMinHits:              " << fcheckMinHits          << std::endl;
    std::cout << "\t CheckContainment:          " << fcheckContainment      << std::endl;
    std::cout << "\t CheckAngle:                " << fcheckAngle            << std::endl;
    std::cout << "\t CheckDirection:            " << fcheckDirection        << std::endl;
    produces< std::vector<novaddt::TriggerDecision>>();
    produces< art::Assns<novaddt::TriggerDecision, novaddt::HitList> >();
}
//////////////////////////////////////////////////////////////////////
novaddt::UpMuTestTrigger::~UpMuTestTrigger()
{
    // Clean up dynamic memory and other resources
}
//////////////////////////////////////////////////////////////////////
bool novaddt::UpMuTestTrigger::filter(art::Event & event)
{
    //////////////////////////////
    //          Events          //
    //////////////////////////////
    
    // UpMuTestTrigger event status
    LOG_DEBUG("UpMuTestTrigger") << "=== novaddt::UpMuTestTrigger filter === :Event: "
    << event.id().event()
    << std::endl;
    _nEvents++;
    
    if(_nEvents % 25 == 0){
        std::cout << "=== novaddt::UpMuTestTrigger === EVENT: " << _nEvents << std::endl;
    }
    
    // Trigger decisions and assignment vectors written out to the event
    std::unique_ptr<std::vector<novaddt::TriggerDecision>> trigger_decisions(new std::vector<novaddt::TriggerDecision>);
    std::unique_ptr<art::Assns<novaddt::TriggerDecision, novaddt::HitList> > assn(new art::Assns<novaddt::TriggerDecision, novaddt::HitList >);
    
    
    
    //////////////////////////////
    //          Tracks          //
    //////////////////////////////
    
    // Obtaining tracks
    art::Handle<std::vector<novaddt::Track> > tracks;
    event.getByLabel(_TrackModuleLabel, _TrackInstanceLabel, tracks);
    LOG_DEBUG("UpMuTestTrigger") << "\tgot " << tracks->size() << " tracks" << std::endl;
    // Obtaining hitlist for each track
    art::FindOneP<novaddt::HitList> fohl(tracks, event, _TrackModuleLabel);
    
    // Examining tracks
    for(size_t i_track = 0; i_track < tracks->size(); ++++i_track){
        _nTracks++;
        novaddt::Track x_track = tracks->at(i_track);
        novaddt::Track y_track = tracks->at(i_track+1);
        // Obtain the hit list for this track
        art::Ptr<novaddt::HitList> x_hit_list = fohl.at(i_track);
        art::Ptr<novaddt::HitList> y_hit_list = fohl.at(i_track+1);
        
        
        
        //////////////////////////////
        //           Hits           //
        //////////////////////////////
        
        /*
         // DEBUG:: Hits loop
         // XZ hits
         for(size_t i_hit = 0; i_hit < x_hit_list->size(); ++i_hit){
         LOG_DEBUG("NuMuTestTrigger") << "\t\t\thit["   << i_hit
         << "]: \tTDC: "   << x_hit_list->at(i_hit).TDC().val
         << ", \tADC: "    << x_hit_list->at(i_hit).ADC().val
         << ", \tplane: "  << x_hit_list->at(i_hit).Plane().val
         << ", \tx-cell: " << x_hit_list->at(i_hit).Cell().val
         << std::endl;
         } // End of loop on x-hits
         
         // YZ hits
         for(size_t i_hit = 0; i_hit < y_hit_list->size(); ++i_hit){
         LOG_DEBUG("NuMuTestTrigger") << "\t\t\thit["   << i_hit
         << "]: \tTDC: "   << y_hit_list->at(i_hit).TDC().val
         << ", \tADC: "    << y_hit_list->at(i_hit).ADC().val
         << ", \tplane: "  << y_hit_list->at(i_hit).Plane().val
         << ", \ty-cell: " << y_hit_list->at(i_hit).Cell().val
         << std::endl;
         } // End of loop on Y-hits
         */
        
        
        
        //////////////////////////////
        //         Trigger          //
        //////////////////////////////
        
        // Checking trigger conditions
        
        //*************************************************//
        //******** Hits cut (minimal track length) ********//
        //*************************************************//
        
        double x_dZ = std::fabs(x_track.EndZ() - x_track.StartZ());
        double y_dZ = std::fabs(y_track.EndZ() - y_track.StartZ());
        double dZ = std::max(x_dZ,y_dZ);
        
        if(fcheckMinHits){
            //CUT::dX or dZ (horizonal projection minimum hits)
            if (x_hit_list->size() < fMinCellsX || dZ < fMinPlanes){
                LOG_DEBUG("UpMuTestTrigger") << "\t\t - failed the trigger decision due to lack horizontal plane hits (dX and dZ cut)" << std::endl;
                continue;
            }
            _nPass_dhorizontal++;
            
            //CUT::dY (vertical projection minimum hits)
            if (y_hit_list->size() < fMinCellsY){
                LOG_DEBUG("UpMuTestTrigger") << "\t\t - failed the trigger decision due to lack of verticle plane hits (dY cut)" << std::endl;
                continue;
            }
            _nPass_dvertical++;
        }
        
        //***************************************************************************//
        //******** Containment cut (through-going, stopped, fully-contained) ********//
        //***************************************************************************//
        
        double min_Z = std::min(x_track.StartZ(),y_track.StartZ());
        double max_Z = std::max(x_track.EndZ(),  y_track.EndZ());
        
        if(fcheckContainment){
            // Through-going muon veto (X and Z horizontal vetos)
            //CUT::X containment
            if ((x_track.StartV() < fMinX) || (x_track.StartV() > fMaxX) ||
                (x_track.EndV()   < fMinX) || (x_track.EndV()   > fMaxX)){
                LOG_DEBUG("UpMuTestTrigger") << "\t\t - failed the trigger decision due to x-containment" << std::endl;
                continue;
            }
            //CUT::Z containment
            if ((min_Z < fMinZ) || (max_Z> fMaxZ) ||
                (max_Z < fMinZ) || (min_Z> fMaxZ)){
                LOG_DEBUG("UpMuTestTrigger") << "\t\t - failed the trigger decision due to z-containment" << std::endl;
                continue;
            }
            _nPass_through++;
            
            // Stopped muon veto (Top Y vertical veto)
            //CUT::Y maximum containment
            if ((y_track.StartV() > fMaxY) || (y_track.EndV()   > fMaxY)){
                LOG_DEBUG("UpMuTestTrigger") << "\t\t - failed the trigger decision due to maximum y-containment" << std::endl;
                continue;
            }
            _nPass_stopped++;
            
            // Fully-contained muon veto (Bottom Y vertical veto)
            //CUT::Y minimum containment
            if ((y_track.StartV() < fMinY) || (y_track.EndV()   < fMinY)){
                LOG_DEBUG("UpMuTestTrigger") << "\t\t - failed the trigger decision due to minimum y-containment" << std::endl;
                continue;
            }
            _nPass_fully++;
        }
        
        //*******************************************//
        //******** Angle cut (vertical cone) ********//
        //*******************************************//
        
        double scaled_dZ = (dZ * 6.6) / 3.9; // Planes longer (z) than Cells (x or y)
        double dX = std::fabs( x_track.EndV() - x_track.StartV() );
        double dY = std::fabs( y_track.EndV() - y_track.StartV() );
        double CosTrack = ( scaled_dZ/ std::sqrt(dX*dX+dY*dY+scaled_dZ*scaled_dZ) );
        
        //CUT::Upward facing 30 degree cone is default
        if (fcheckAngle){
            if (CosTrack > fAngleCut){
                LOG_DEBUG("UpMuTestTrigger") << "\t\t - failed the trigger decision due to angle cut" << std::endl;
                continue;
            }
            _nPass_angle++;
        }
        
        /*
         //// EXTRA::Alternative angular cut-- a simple loose cut on the inverse slope |dz|/|dy|
         //// inspired by doc-db 8888
         //// at the moment I'm not sure if the x, y & z coordinates provided are sensible
         //// and in the same aspect ratio.
         //// I doubt they are. This should be re-visited
         //double dY = std::fabs( y_track.EndV() - y_track.StartV() );
         ////double dZ = std::fabs( tracks->at(i_track).EndZ() - tracks->at(i_track).StartZ() );
         //double inverse_slope = dZ / dY;
         //std::cout << "\t\t\tdY: "        << dY
         //<< ", dZ: "            << dZ
         //<< ", inverse slope: " << inverse_slope
         //<< std::endl;
         //if (inverse_slope > _inverseSlopeCut){
         //std::cout << "\t\t - failed the trigger decision due to inverse slope cut" << std::endl;
         //continue;
         //}
         */
        
        //**************************************************//
        //******** Direction cut (upward traveling) ********//
        //**************************************************//
        
        if(fcheckDirection){
            /*
             // Uncertainty analysis
             double min_tdc = std::min(x_hit_list->front().TDC().val,y_hit_list->front().TDC().val);
             double max_tdc = std::max(x_hit_list->back().TDC().val, y_hit_list->back().TDC().val);
             double dTime = max_tdc - min_tdc; // Time separation between track beginning and end
             double dLength = std::sqrt(dX*dX+dY*dY+scaled_dZ*scaled_dZ); // Track length
             */
            
            //################# ALGORITHM 1 #################//
            //          Slope correlation method           ///
            //###############################################//
            /*
             // Use T for Y cells
             
             // Sort hits by Y cell position
             unsigned int nHits = y_hit_list->size();
             //art::Handle<novaddt::HitList> hits;
             HitList hit_list(*y_hit_list);
             std::sort(hit_list.begin(), hit_list.end(),CompareDAQHit<Cell>());
             
             // Definitions
             int loop = 0;
             unsigned int nPairs = (nHits+1) * nHits / 2;
             double tmte=0, tm=0, te=0, te2=0, slp, angle;
             double mea[nPairs], exp[nPairs];
             
             for(size_t i_hit = 0; i_hit < nHits; i_hit++){
             for (size_t j_hit = 0; j_hit < i_hit; j_hit++){
             // Variables
             //Ti = (hit_list.at(i_hit).TDC().val * 15.625) - ((-hit_list.at(i_hit).Cell().val +  192) * dC / 1.59); // Photon transport (fiber optical) corrected time
             //Tj = (hit_list.at(j_hit).TDC().val * 15.625) - ((-hit_list.at(j_hit).Cell().val +  192) * dC / 1.59); // n = 1.59 index; (dx/2) = 192 cells;
             
             // Measured time
             mea[loop] = hit_list.at(i_hit).TDC().val*15.625 - hit_list.at(j_hit).TDC().val*15.625; // TDC->ns conversion (64 MHz clock ticks)
             fmeasured->Fill(mea[loop]);
             
             // Expected time
             exp[loop] = hit_list.at(i_hit).Cell().val*0.13008999712728 - hit_list.at(j_hit).Cell().val*0.13008999712728; // Cell->ns converstion (3.9*1e7/299792458)
             fexpected->Fill(exp[loop]);
             
             // Analysis
             fexpmea->Fill(exp[loop], mea[loop]);
             
             // Counters
             tmte += mea[loop] * exp[loop];
             tm += mea[loop];
             te += exp[loop];
             te2 += exp[loop] * exp[loop];
             loop++;
             
             // Check timing (j)
             //std::cout << "Expected: " << exp[loop] << ", Measured: "  <<  mea[loop]   <<  std::endl;
             }  // End of x comp hits loops (j_hit)
             
             // Check timing (i)
             //std::cout << Ti << std::endl;
             } // End of x test hits loop (i_hit)
             
             // Metrics
             slp = (loop*tmte-tm*te)/(loop*te2-te*te);
             //angle = std::acos(dY/scaled_dZ) * 180 / 3.14159;
             angle = std::acos(CosTrack) * 180 / 3.14159;
             //std::cout << "Track time sigma angle: " << angle << ",  Track time sigma slope: "  <<  slp <<  std::endl;
             
             fslope->Fill(slp);
             fdirslp->Fill(slp, angle);
             */
            //################# ALGORITHM 2 #################//
            ///           Adjacent cell analysis            ///
            //###############################################//
            
            // Get Y position for X cells-- use X cell's T for these for better resolution (add in Y cell's T too?) (use mergetracks/Track3D to get vertex?)
            
            unsigned int yHits = y_hit_list->size();
            HitList check_hit_list(*y_hit_list); // Y hitlist
            unsigned int xHits = x_hit_list->size();
            HitList hit_list(*x_hit_list); // X hitlist
            
            // Sort X hits by Y cell position
            std::sort(check_hit_list.begin(), check_hit_list.end(), CompareDAQHit<Plane>());
            int sortY = check_hit_list.at(yHits-1).Cell().val - check_hit_list.at(0).Cell().val; // pos for +y, nega for -y:: .back() - .front()
            if(sortY < 0){
                std::sort(hit_list.rbegin(), hit_list.rend(), CompareDAQHit<Plane>()); // descending (reverse) order
                //std::sort(hit_list.begin(), hit_list.end(), CompareDAQHit<Plane>());
                //std::reverse(hit_list.begin(),hit_list.end());
            } else{
                std::sort(hit_list.begin(), hit_list.end(), CompareDAQHit<Plane>()); // ascending (normal) order
            }
            
            //            // Get Y position for X cells (relative to first cell)
            //            //std::sort(check_hit_list.begin(), check_hit_list.end(), CompareDAQHit<Cell>());
            //            double dYc, dYp, sum=0, m;
            //            for(size_t y_hit = 0; y_hit < (yHits - 1); y_hit++){
            //                dYc= check_hit_list.at(y_hit+1).Cell().val - check_hit_list.at(y_hit).Cell().val;
            //                dYp= check_hit_list.at(y_hit+1).Plane().val - check_hit_list.at(y_hit).Plane().val;
            //                sum += dYc/dYp;
            //                std::cout << sum << std::endl;
            //
            //            }
            
            // Get Y position for X cells (relative to first cell) [need to improve slope determination with more points]
            double dYc, dYp, m;
            dYc= check_hit_list.at(yHits-1).Cell().val - check_hit_list.at(0).Cell().val;
            dYp= check_hit_list.at(yHits-1).Plane().val - check_hit_list.at(0).Plane().val;
            m = dYc / dYp; // mean slope of Y cell positions as a function of plane (take average of y and see if negative or positive-- can be used for sorting instead of current method)
            
            // Fill Y position for sorted X cells
            double AltCell[xHits];  // Y position of X cells relative to first X cell
            for(size_t x_hit = 0; x_hit < xHits; x_hit++){
                AltCell[x_hit] = m * hit_list.at(x_hit).Plane().val + m * hit_list.at(0).Plane().val;
            }
            
            //             // Sort hits by Y cell position
            //             if(m > 0){
            //             std::sort(hit_list.begin(), hit_list.end(), CompareDAQHit<Cell>()); // ascending order (Plane?)
            //             } else{
            //             std::sort(hit_list.begin(), hit_list.end(), CompareDAQHit<Cell>());
            //             std::reverse(hit_list.begin(),hit_list.end());
            //             //std::sort(hit_list.rbegin(), hit_list.rend(), CompareDAQHit<Cell>()); // descending (reverse) order
            //             }
            
            //
            //             // Sort hits by Y cell position
            //             for(size_t i_hit = 0; i_hit < nHits; i_hit++){
            //             for (size_t j_hit = 0; j_hit < i_hit; j_hit++){
            //             if(AltCell[j_hit] < AltCell[i_hit]){
            //             DAQHit tmp = hit_list.at(i_hit);
            //             double tmpalt = AltCell[i_hit];
            //
            //             hit_list.at(i_hit) = hit_list.at(j_hit);
            //             AltCell[i_hit] = AltCell[j_hit];
            //
            //             hit_list.at(j_hit) = tmp;
            //             AltCell[j_hit] = tmpalt;
            //             }
            //             }
            //             }
            
            // Definitions
            unsigned int nPairs = (xHits+1) * xHits / 2;
            double tmte=0, tm=0, te=0, te2=0, slp, angle, dT, dC;
            double mea[nPairs], exp[nPairs];
            int loop = 0;
            
            for(size_t i_hit = 0; i_hit < xHits; i_hit++){
                for (size_t j_hit = 0; j_hit < i_hit; j_hit++){
                    // Variables
                    //Ti = (hit_list.at(i_hit).TDC().val * 15.625) - ((-hit_list.at(i_hit).Cell().val +  192) * dC / 1.59); // Photon transport (fiber optical) corrected time
                    //Tj = (hit_list.at(j_hit).TDC().val * 15.625) - ((-hit_list.at(j_hit).Cell().val +  192) * dC / 1.59); // n = 1.59 index; (dx/2) = 192 cells;
                    dT = hit_list.at(i_hit).TDC().val * 15.625 - hit_list.at(j_hit).TDC().val * 15.625; // TDC->ns conversion (64 MHz clock ticks)
                    //dC = hit_list.at(i_hit).Cell().val * 0.13008999712728 - hit_list.at(j_hit).Cell().val * 0.13008999712728; // Cell->ns converstion (speed of light and cell scaling-- 3.9*1e7/299792458) // X position
                    dC = AltCell[i_hit] * 0.13008999712728 - AltCell[j_hit] * 0.13008999712728; // Y position
                    
                    // Measured time
                    mea[loop] = dT;
                    fmeasured->Fill(mea[loop]);
                    
                    // Expected time
                    exp[loop] = dC;
                    fexpected->Fill(exp[loop]);
                    
                    // Analysis
                    fexpmea->Fill(exp[loop], mea[loop]);
                    
                    // Counters
                    tmte += mea[loop] * exp[loop];
                    tm += mea[loop];
                    te += exp[loop];
                    te2 += exp[loop] * exp[loop];
                    loop++;
                    
                    // Check timing (j)
                    //std::cout << "Expected: " << exp[loop] << ", Measured: "  <<  mea[loop]   <<  std::endl;
                }  // End of x comp hits loops (j_hit)
                
                // Check timing (i)
                //std::cout << Ti << std::endl;
            } // End of x test hits loop (i_hit)
            
            // Metrics
            slp = (loop*tmte-tm*te)/(loop*te2-te*te);
            //angle = std::acos(dY/scaled_dZ) * 180 / 3.14159;
            angle = std::acos(CosTrack) * 180 / 3.14159;
            //std::cout << "Track time sigma angle: " << angle << ",  Track time sigma slope: "  <<  slp <<  std::endl;
            
            fslope->Fill(slp);
            fdirslp->Fill(slp, angle);
            
            //---------------------------------------------------------------------------
            // CUT::Slope must be positive to be upward going (assuming fDirectionCut==0)
            if(slp < fDirectionCut){
                LOG_DEBUG("UpMuTestTrigger") << "\t\t - failed the trigger decision due to direction cut" << std::endl;
                continue;
            }
            _nPass_direction++;
        }
        
        
        
        //////////////////////////////
        //           UpMu           //
        //////////////////////////////
        
        // Track has passed trigger decision and is now identified as upward traveling muon
        _trigger_counts++;
        // check the prescale
        if (_trigger_counts%_prescale == _prescale-1) {
            _prescaled_trigger_counts++;
            LOG_DEBUG("NuMuTrigger") << "\t\t - not prescaled" << std::endl;
            //DEBUG::Analyze UpMu tracks
            // XZ projection
            for(size_t i_hit = 0; i_hit < x_hit_list->size(); ++i_hit){
                HitsXZ -> Fill(x_hit_list->at(i_hit).Plane().val, x_hit_list->at(i_hit).Cell().val);
            } // End of loop on x hits
            // YZ projection
            for(size_t i_hit = 0; i_hit < y_hit_list->size(); ++i_hit){
                HitsYZ -> Fill(y_hit_list->at(i_hit).Plane().val, y_hit_list->at(i_hit).Cell().val);
            } // End of loop on x hits
            
            uint64_t min_tdc_trig = std::min(x_hit_list->front().TDC().val,y_hit_list->front().TDC().val);
            uint64_t max_tdc_trig = std::max(x_hit_list->back().TDC().val, y_hit_list->back().TDC().val);
            trigger_decisions->push_back(novaddt::TriggerDecision(
                                                                  min_tdc_trig,
                                                                  max_tdc_trig - min_tdc_trig,
                                                                  novaddt::TriggerMask(daqdataformats::trigID::TRIG_ID_DATA_UPMU),
                                                                  _prescale)
                                         );
            util::CreateAssn(*this, event, *trigger_decisions, x_hit_list, *assn);
            util::CreateAssn(*this, event, *trigger_decisions, y_hit_list, *assn);
            LOG_DEBUG("UpMuTestTrigger") << "\t\t - passed the trigger decision, start time: "
            << trigger_decisions->back().start()
            << ", duration: " << trigger_decisions->back().duration()
            << std::endl;
        } // End of check on prescale
    } // End of loop on tracks
    
    bool goodTrigger = (trigger_decisions->size() > 0);
    LOG_DEBUG("UpMuTestTrigger") << "=== end of novaddt::UpMuTestTrigger filter === :Trigger decision: " << goodTrigger
    << " from " << trigger_decisions->size()
    << " passes"<< std::endl;
    
    event.put(std::move(trigger_decisions));
    event.put(std::move(assn));
    
    return goodTrigger;
} // End of loop on events
//////////////////////////////////////////////////////////////////////
void novaddt::UpMuTestTrigger::endJob()
{
    std::cout << "=== novaddt::UpMuTestTrigger end ===              " << std::endl;
    std::cout << "\tNumber of events:                               " << _nEvents             << std::endl;
    std::cout << "\tNumber of tracks:                               " << _nTracks             << " / " << "Trigger Rejection" << std::endl;
    std::cout << "\tNumber that pass min horizontal hits:           " << _nPass_dhorizontal   << " / " << 100 - (100*(float)(_nPass_dhorizontal)/(float)(_nTracks)) << "%" << std::endl;
    std::cout << "\tNumber that pass min vertical hits:             " << _nPass_dvertical     << " / " << 100 - (100*(float)(_nPass_dvertical)/(float)(_nTracks)) << "%" << std::endl;
    std::cout << "\tNumber that pass through containment (X/Z):     " << _nPass_through       << " / " << 100 - (100*(float)(_nPass_through)/(float)(_nTracks)) << "%" << std::endl;
    std::cout << "\tNumber that pass stopped containment (MaxY):    " << _nPass_stopped       << " / " << 100 - (100*(float)(_nPass_stopped)/(float)(_nTracks)) << "%" << std::endl;
    std::cout << "\tNumber that pass fully containment (MinY):      " << _nPass_fully         << " / " << 100 - (100*(float)(_nPass_fully)/(float)(_nTracks)) << "%" << std::endl;
    std::cout << "\tNumber that pass angle:                         " << _nPass_angle         << " / " << 100 - (100*(float)(_nPass_angle)/(float)(_nTracks)) << "%" << std::endl;
    std::cout << "\tNumber that pass direction:                     " << _nPass_direction     << " / " << 100 - (100*(float)(_nPass_direction)/(float)(_nTracks)) << "%" << std::endl;
    std::cout << "\tNumber that pass the trigger:                   " << _trigger_counts      << std::endl;
    std::cout << "\t              after prescale:                   " << _prescaled_trigger_counts << std::endl;
    // Write analysis histograms
    TFile f("upmu_hists.root","recreate");
    
    HitsXZ->Write();
    HitsYZ->Write();
    
    fslope->Write();
    fexpected->Write();
    fmeasured->Write();
    fdirslp->Write();
    fexpmea->Write();
    /*
     ftrue->Write();
     fsmear->Write();
     fupmuper->Write();
     */
}

//////////////////////////////////////////////////////////////////////
DEFINE_ART_MODULE(novaddt::UpMuTestTrigger)
