#ifndef DDT_HOUGH_TRACK_MAKER_HH_GUARD
#define DDT_HOUGH_TRACK_MAKER_HH_GUARD

#include <art/Framework/Core/EDFilter.h>
#include <art/Framework/Principal/Event.h>
#include <art/Persistency/Common/Assns.h>
#include <fhiclcpp/ParameterSet.h>

#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/HoughPoint.h"
#include "DDTBaseDataProducts/Track.h"

#include <tbb/tbb.h>

#include <vector>

namespace novaddt
{
  typedef tbb::blocked_range<art::PtrVector<HitList>::const_iterator> br_t;
  class HoughTrackMaker;
  struct hitcomp;
}


class novaddt::HoughTrackMaker
{
public:
  HoughTrackMaker() = default;
  HoughTrackMaker(fhicl::ParameterSet const& p);

  HoughTrackMaker(HoughTrackMaker const& other, tbb::split);
  ~HoughTrackMaker();

  // for TBB parallelization:
  void operator()(br_t const& slices);
  void join(HoughTrackMaker const& other);

  // standard serial way:
  void do_serial(art::PtrVector<HitList> const& slices);

  void create_associations
  (art::EDFilter const& filter, 
   art::Event & e, 
   art::Assns<novaddt::Track, novaddt::HitList> & assn,
   art::Assns<novaddt::Track, novaddt::HitList> & assn_b) const;
  void move_collections_into_event(art::Event & e);


private:
  bool PairCutter
  (HitList::const_iterator hit_ptr1, HitList::const_iterator hit_ptr2) const;
  bool do_like_zukai(art::Ptr<HitList> slice);

  // FHICL Parameters:
  std::string     _sliceLabel;
  std::string     _sliceInstance;
  uint64_t        _tdcSigma;
  int             _planeSigma;
  int             _cellSigma;
  float           _vCut;
  unsigned int    _minPointsToTag;
  unsigned int    _minHitsPerTrack;

  unsigned int    _maxHitsInSlice; //! Maximum Number of hits in a slice that will be processes

  static const int fcGrid_ = 100;
  static const int fdGrid_ = 100;

  bool flag_[2][fcGrid_][fdGrid_];
  unsigned pool_location_[2][fcGrid_][fdGrid_];
  std::vector<HoughPoint*> pair_pool_[2][fcGrid_][fdGrid_];

  std::unique_ptr<std::vector<Track> > tracks_;
  std::unique_ptr<std::vector<HitList> > track_hits_;

  std::vector<art::Ptr<HitList> > track_slice_;
};


struct novaddt::hitcomp 
{
  bool operator() (const novaddt::DAQHit* lhs, const novaddt::DAQHit* rhs) const
  {
    if (!(lhs->TDC().val == rhs->TDC().val))
      return lhs->TDC().val < rhs->TDC().val;
    else if (!(lhs->Plane() == rhs->Plane())) 
      return lhs->Plane() < rhs->Plane();
    else
      return lhs->Cell() < rhs->Cell();
  }
};


#endif
